unit jifformu;

// Copyright  1999 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls;

type
  TJifForm = class(TForm)
    cbCurrent : TCheckBox;
    cbDefault : TCheckBox;
    cbAll     : TCheckBox;
    ebJifs    : TEdit;
    Label1    : TLabel;
    udJifs    : TUpDown;
    btnHelp   : TButton;
    btnCancel : TButton;
    btnOK     : TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbSharedClick(Sender: TObject);
    procedure ebJifsKeyPress(Sender: TObject; var Key: Char);
    procedure ebJifsChange(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure SetData(DefOnly : Boolean; allJifs, J : Integer);
  end;

var
  JifForm: TJifForm;

implementation

uses allfuncs;

{$R *.DFM}

procedure TJifForm.FormCreate(Sender: TObject);
begin
  GetPosFmExeName(Application.ExeName, Self, False);
end;

procedure TJifForm.FormDestroy(Sender: TObject);
begin
  SetPosToExeName(Application.ExeName, Self, False);
end;

procedure TJifForm.cbSharedClick(Sender: TObject);
begin
  btnOK.Enabled := cbCurrent.Checked OR cbDefault.Checked OR
    cbAll.Checked;
end;

procedure TJifForm.ebJifsKeyPress(Sender: TObject; var Key: Char);
// Only accept ctrl-keys or up-to-three digits
begin
  CASE Key OF
    #0..#31  : ; //OK
    '0'..'9' : IF (Length(ebJifs.Text) >= 3) AND
      (ebJifs.SelLength = 0) THEN Key := #0;
    ELSE Key := #0;
  end;
end;

procedure TJifForm.ebJifsChange(Sender: TObject);
// Strip non-digits from pasted-in text and limit to 3 digits
VAR
  S : String;
  N : Integer;
begin
  S := (Sender AS TEdit).Text;
  FOR N := Length(S) DOWNTO 1 DO
    CASE S[N] OF
      '0'..'9' : ; // OK
      ELSE Delete(S, N, 1);
    end;
  IF Length(S) > 3 THEN SetLength(S, 3);  
  IF S = (Sender AS TEdit).Text THEN Exit;
  WITH Sender AS TEdit DO
    begin
      Text      := S;
      SelStart  := 0;
      SelLength := Length(S);
    end;
end;

procedure TJifForm.btnHelpClick(Sender: TObject);
begin
  Application.HelpContext(2);
end;

procedure TJifForm.SetData(DefOnly : Boolean; allJifs, J : Integer);
begin
  cbDefault.Caption := Format(cbDefault.Caption, [allJifs]);
  udJifs.Position   := J;
  cbDefault.Enabled := True;
  cbAll.Checked     := False;
  IF DefOnly THEN
    begin
      cbDefault.Checked := True;
      cbCurrent.Enabled := False;
      cbCurrent.Checked := False;
      cbAll.Enabled     := False;
    end
  ELSE
    begin
      cbDefault.Checked := False;
      cbCurrent.Enabled := True;
      cbCurrent.Checked := True;
      cbAll.Enabled     := True;
    end;
end;

end.



